from copy import copy, deepcopy
from datetime import datetime
from uuid import uuid4

from pytest import raises

from haiway.utils.metadata import META_EMPTY, Meta, _validated_meta_value


def test_meta_empty_is_singleton():
    # META_EMPTY is only returned for None, not empty dict
    assert Meta.of(None) is META_EMPTY
    assert not META_EMPTY
    assert len(META_EMPTY) == 0

    # Empty dict creates a new instance (but still empty)
    empty_from_dict = Meta.of({})
    assert not empty_from_dict
    assert len(empty_from_dict) == 0


def test_meta_basic_construction():
    meta = Meta({"kind": "test", "name": "example"})
    assert meta["kind"] == "test"
    assert meta["name"] == "example"
    assert len(meta) == 2
    assert bool(meta) is True


def test_meta_construction_validates_values():
    # Direct constructor doesn't validate - use of() or from_mapping()
    meta = Meta.from_mapping({"tags": ["a", "b"], "nested": {"key": "value"}})
    assert meta["tags"] == ("a", "b")  # converted to tuple
    assert meta["nested"] == {"key": "value"}


def test_meta_of_with_none_returns_empty():
    result = Meta.of(None)
    assert result is META_EMPTY


def test_meta_of_with_existing_meta_returns_same():
    original = Meta({"kind": "test"})
    result = Meta.of(original)
    assert result is original


def test_meta_of_with_mapping_validates():
    result = Meta.of({"tags": ["a", "b"], "active": True})
    assert result["tags"] == ("a", "b")
    assert result["active"] is True


def test_meta_from_mapping():
    result = Meta.from_mapping({"kind": "user", "count": 42})
    assert result.kind == "user"
    assert result["count"] == 42


def test_meta_from_json_valid():
    json_data = '{"kind": "test", "tags": ["a", "b"], "active": true}'
    result = Meta.from_json(json_data)
    assert result.kind == "test"
    assert result["tags"] == ("a", "b")
    assert result["active"] is True


def test_meta_from_json_invalid():
    with raises(ValueError, match="Invalid json"):
        Meta.from_json('"not an object"')


def test_meta_to_json():
    meta = Meta({"kind": "test", "active": True})
    json_str = meta.to_json()
    assert "kind" in json_str
    assert "test" in json_str


def test_meta_to_mapping():
    original_mapping = {"kind": "test", "active": True}
    meta = Meta(original_mapping)
    result = meta.to_mapping()
    assert result == original_mapping


def test_meta_kind_property():
    meta = Meta({"kind": "user"})
    assert meta.kind == "user"

    empty_meta = Meta({})
    assert empty_meta.kind is None

    invalid_meta = Meta({"kind": 123})
    assert invalid_meta.kind is None


def test_meta_with_kind():
    meta = Meta({"name": "test"})
    updated = meta.with_kind("user")
    assert updated.kind == "user"
    assert updated["name"] == "test"
    assert meta.kind is None  # original unchanged


def test_meta_name_property():
    meta = Meta({"name": "John"})
    assert meta.name == "John"

    empty_meta = Meta({})
    assert empty_meta.name is None


def test_meta_with_name():
    meta = Meta({"kind": "user"})
    updated = meta.with_name("Jane")
    assert updated.name == "Jane"
    assert updated.kind == "user"


def test_meta_description_property():
    meta = Meta({"description": "A test object"})
    assert meta.description == "A test object"

    empty_meta = Meta({})
    assert empty_meta.description is None


def test_meta_with_description():
    meta = Meta({"name": "test"})
    updated = meta.with_description("Test description")
    assert updated.description == "Test description"
    assert updated.name == "test"


def test_meta_identifier_property():
    test_uuid = uuid4()
    meta = Meta({"identifier": str(test_uuid)})
    assert meta.identifier == test_uuid

    empty_meta = Meta({})
    assert empty_meta.identifier is None

    invalid_meta = Meta({"identifier": "not-a-uuid"})
    assert invalid_meta.identifier is None


def test_meta_with_identifier():
    test_uuid = uuid4()
    meta = Meta({"name": "test"})
    updated = meta.with_identifier(test_uuid)
    assert updated.identifier == test_uuid
    assert updated["identifier"] == str(test_uuid)


def test_meta_origin_identifier_property():
    test_uuid = uuid4()
    meta = Meta({"origin_identifier": str(test_uuid)})
    assert meta.origin_identifier == test_uuid


def test_meta_with_origin_identifier():
    test_uuid = uuid4()
    meta = Meta({})
    updated = meta.with_origin_identifier(test_uuid)
    assert updated.origin_identifier == test_uuid


def test_meta_predecessor_identifier_property():
    test_uuid = uuid4()
    meta = Meta({"predecessor_identifier": str(test_uuid)})
    assert meta.predecessor_identifier == test_uuid


def test_meta_with_predecessor_identifier():
    test_uuid = uuid4()
    meta = Meta({})
    updated = meta.with_predecessor_identifier(test_uuid)
    assert updated.predecessor_identifier == test_uuid


def test_meta_successor_identifier_property():
    test_uuid = uuid4()
    meta = Meta({"successor_identifier": str(test_uuid)})
    assert meta.successor_identifier == test_uuid


def test_meta_with_successor_identifier():
    test_uuid = uuid4()
    meta = Meta({})
    updated = meta.with_successor_identifier(test_uuid)
    assert updated.successor_identifier == test_uuid


def test_meta_tags_property():
    meta = Meta({"tags": ["active", "verified"]})
    assert meta.tags == ("active", "verified")

    empty_meta = Meta({})
    assert empty_meta.tags == ()

    mixed_meta = Meta({"tags": ["valid", 123, "also_valid"]})
    assert mixed_meta.tags == ("valid", "also_valid")  # filters non-strings


def test_meta_with_tags_new():
    meta = Meta({})
    updated = meta.with_tags(["new", "tags"])
    assert updated.tags == ("new", "tags")


def test_meta_with_tags_append():
    meta = Meta({"tags": ["existing"]})
    updated = meta.with_tags(["new", "existing"])  # existing should be deduplicated
    assert "existing" in updated.tags
    assert "new" in updated.tags


def test_meta_has_tags():
    meta = Meta({"tags": ["active", "verified", "premium"]})
    assert meta.has_tags(["active"]) is True
    assert meta.has_tags(["active", "verified"]) is True
    assert meta.has_tags(["nonexistent"]) is False
    assert meta.has_tags(["active", "nonexistent"]) is False


def test_meta_creation_property():
    now = datetime.now()
    meta = Meta({"creation": now.isoformat()})
    # Compare by converting both back to isoformat to handle precision
    assert meta.creation.isoformat() == now.isoformat()

    empty_meta = Meta({})
    assert empty_meta.creation is None

    invalid_meta = Meta({"creation": "not-a-date"})
    assert invalid_meta.creation is None


def test_meta_with_creation():
    now = datetime.now()
    meta = Meta({})
    updated = meta.with_creation(now)
    assert updated.creation is not None
    assert updated["creation"] == now.isoformat()


def test_meta_merged_with():
    original = Meta({"name": "original", "kind": "test"})
    updates = {"name": "updated", "description": "merged"}

    merged = original.merged_with(updates)
    assert merged["name"] == "updated"
    assert merged["kind"] == "test"  # preserved
    assert merged["description"] == "merged"  # added


def test_meta_merged_with_none():
    original = Meta({"name": "test"})
    result = original.merged_with(None)
    assert result is original  # same instance returned


def test_meta_merged_with_meta_instance():
    original = Meta({"name": "original"})
    other = Meta({"name": "updated", "kind": "test"})

    merged = original.merged_with(other)
    assert merged["name"] == "updated"
    assert merged["kind"] == "test"


def test_meta_excluding():
    original = Meta({"name": "test", "kind": "user", "description": "desc"})

    result = original.excluding("kind", "description")
    assert "name" in result
    assert "kind" not in result
    assert "description" not in result


def test_meta_excluding_none():
    original = Meta({"name": "test"})
    result = original.excluding()
    assert result is original  # same instance


def test_meta_updated():
    original = Meta({"name": "original"})
    updated = original.updated(name="updated", kind="test")
    assert updated["name"] == "updated"
    assert updated["kind"] == "test"


def test_meta_immutability():
    meta = Meta({"name": "test"})

    with raises(AttributeError, match="Can't modify immutable"):
        meta.name = "changed"  # type: ignore

    with raises(AttributeError, match="Can't modify immutable"):
        del meta.name  # type: ignore

    with raises(AttributeError, match="Can't modify immutable"):
        meta["name"] = "changed"  # type: ignore

    with raises(AttributeError, match="Can't modify immutable"):
        del meta["name"]  # type: ignore


def test_meta_mapping_interface():
    meta = Meta({"a": 1, "b": 2, "c": 3})

    # Test iteration
    keys = list(meta)
    assert set(keys) == {"a", "b", "c"}

    # Test containment
    assert "a" in meta
    assert "d" not in meta

    # Test get method (inherited from Mapping)
    assert meta.get("a") == 1
    assert meta.get("d") is None
    assert meta.get("d", "default") == "default"


def test_meta_copy_returns_same_instance():
    meta = Meta({"name": "test"})
    assert copy(meta) is meta
    assert deepcopy(meta) is meta


def test__validated_meta_value_primitives():
    assert _validated_meta_value(None) is None
    assert _validated_meta_value("string") == "string"
    assert _validated_meta_value(42) == 42
    assert _validated_meta_value(3.14) == 3.14
    assert _validated_meta_value(True) is True
    assert _validated_meta_value(False) is False


def test__validated_meta_value_collections():
    # Lists become tuples
    assert _validated_meta_value([1, 2, 3]) == (1, 2, 3)

    # Nested validation
    result = _validated_meta_value([1, [2, 3], {"a": "b"}])
    assert result == (1, (2, 3), {"a": "b"})

    # Dicts are validated recursively
    result = _validated_meta_value({"list": [1, 2], "nested": {"key": "value"}})
    assert result == {"list": (1, 2), "nested": {"key": "value"}}


def test__validated_meta_value_invalid():
    class CustomObject:
        pass

    with raises(TypeError, match="Invalid Meta value"):
        _validated_meta_value(CustomObject())

    with raises(TypeError, match="Invalid Meta value"):
        _validated_meta_value(object())



import json
from collections.abc import Collection, Iterator, Mapping
from datetime import datetime
from types import EllipsisType
from typing import Any, ClassVar, Final, Self, cast, final
from uuid import UUID

__all__ = (
    "META_EMPTY",
    "Meta",
    "MetaTags",
    "MetaValue",
    "MetaValues",
)

type MetaValue = Mapping[str, MetaValue] | Collection[MetaValue] | str | float | int | bool | None
type MetaValues = Mapping[str, MetaValue]
type MetaTags = Collection[str]


@final
class Meta(Mapping[str, MetaValue]):
    """
    Immutable metadata container with type-safe access to common fields.

    Meta provides a structured way to store and access metadata as key-value pairs
    with built-in support for common metadata fields like identifiers, names,
    descriptions, tags, and timestamps. All values are validated to ensure they
    conform to the MetaValue type constraints.

    The class implements the Mapping protocol, allowing dict-like access while
    maintaining immutability. It provides builder-style methods (with_*) for
    creating modified copies and convenience properties for accessing standard
    metadata fields.

    Examples
    --------
    >>> meta = Meta.of({"kind": "user", "name": "John"})
    >>> meta = meta.with_tags(["active", "verified"])
    >>> print(meta.kind)  # "user"
    >>> print(meta.tags)  # ("active", "verified")
    """

    __IMMUTABLE__: ClassVar[EllipsisType] = ...

    __slots__ = ("_values",)

    def __init__(
        self,
        values: MetaValues,
        /,
    ):
        assert isinstance(values, Mapping)  # nosec: B101
        self._values: MetaValues
        object.__setattr__(
            self,
            "_values",
            values,
        )

    @classmethod
    def of(
        cls,
        meta: Self | MetaValues | None,
    ) -> Self:
        """
        Create a Meta instance from various input types.

        This factory method provides a flexible way to create Meta instances,
        handling None values, existing Meta instances, and raw mappings.

        Parameters
        ----------
        meta : Self | MetaValues | None
            The metadata to wrap. Can be None (returns META_EMPTY),
            an existing Meta instance (returns as-is), or a mapping
            of values to validate and wrap.

        Returns
        -------
        Self
            A Meta instance containing the provided metadata.
        """
        if meta is None:
            return cast(Self, META_EMPTY)

        elif isinstance(meta, Meta):
            return cast(Self, meta)

        else:
            assert isinstance(meta, Mapping)  # nosec: B101
            return cls({key: _validated_meta_value(value) for key, value in meta.items()})

    @classmethod
    def from_mapping(
        cls,
        mapping: Mapping[str, Any],
        /,
    ) -> Self:
        return cls({key: _validated_meta_value(value) for key, value in mapping.items()})

    @classmethod
    def from_json(
        cls,
        value: str | bytes,
        /,
    ) -> Self:
        match json.loads(value):
            case {**values}:
                return cls({key: _validated_meta_value(val) for key, val in values.items()})

            case other:
                raise ValueError(f"Invalid json: {other}")

    def to_str(self) -> str:
        return self.__str__()

    def to_mapping(
        self,
    ) -> Mapping[str, Any]:
        return self._values

    def to_json(
        self,
    ) -> str:
        return json.dumps(self._values)

    @property
    def kind(self) -> str | None:
        match self._values.get("kind"):
            case str() as kind:
                return kind

            case _:
                return None

    def with_kind(
        self,
        kind: str,
        /,
    ) -> Self:
        return self.__class__(
            {
                **self._values,
                "kind": kind,
            }
        )

    def _get_uuid(
        self,
        key: str,
        /,
    ) -> UUID | None:
        match self._values.get(key):
            case str() as identifier:
                try:
                    return UUID(identifier)
                except ValueError:
                    return None
            case _:
                return None

    def _with_uuid(
        self,
        key: str,
        /,
        *,
        value: UUID,
    ) -> Self:
        return self.__class__(
            {
                **self._values,
                key: str(value),
            }
        )

    @property
    def identifier(self) -> UUID | None:
        return self._get_uuid("identifier")

    def with_identifier(
        self,
        identifier: UUID,
        /,
    ) -> Self:
        return self._with_uuid(
            "identifier",
            value=identifier,
        )

    @property
    def origin_identifier(self) -> UUID | None:
        return self._get_uuid("origin_identifier")

    def with_origin_identifier(
        self,
        identifier: UUID,
        /,
    ) -> Self:
        return self._with_uuid(
            "origin_identifier",
            value=identifier,
        )

    @property
    def predecessor_identifier(self) -> UUID | None:
        return self._get_uuid("predecessor_identifier")

    def with_predecessor_identifier(
        self,
        identifier: UUID,
        /,
    ) -> Self:
        return self._with_uuid(
            "predecessor_identifier",
            value=identifier,
        )

    @property
    def successor_identifier(self) -> UUID | None:
        return self._get_uuid("successor_identifier")

    def with_successor_identifier(
        self,
        identifier: UUID,
        /,
    ) -> Self:
        return self._with_uuid(
            "successor_identifier",
            value=identifier,
        )

    @property
    def name(self) -> str | None:
        match self._values.get("name"):
            case str() as name:
                return name

            case _:
                return None

    def with_name(
        self,
        name: str,
        /,
    ) -> Self:
        return self.__class__(
            {
                **self._values,
                "name": name,
            }
        )

    @property
    def description(self) -> str | None:
        match self._values.get("description"):
            case str() as description:
                return description

            case _:
                return None

    def with_description(
        self,
        description: str,
        /,
    ) -> Self:
        return self.__class__(
            {
                **self._values,
                "description": description,
            }
        )

    @property
    def tags(self) -> MetaTags:
        match self._values.get("tags"):
            case [*tags]:
                return tuple(tag for tag in tags if isinstance(tag, str))

            case _:
                return ()

    def with_tags(
        self,
        tags: MetaTags,
        /,
    ) -> Self:
        match self._values.get("tags"):
            case [*current_tags]:
                return self.__class__(
                    {
                        **self._values,
                        "tags": (
                            *current_tags,
                            *(
                                _validated_meta_value(tag)
                                for tag in tags
                                if tag not in current_tags
                            ),
                        ),
                    }
                )

            case _:
                return self.__class__({**self._values, "tags": tags})

    def has_tags(
        self,
        tags: MetaTags,
        /,
    ) -> bool:
        match self._values.get("tags"):
            case [*meta_tags]:
                return all(tag in meta_tags for tag in tags)

            case _:
                return False

    @property
    def creation(self) -> datetime | None:
        match self._values.get("creation"):
            case str() as iso_value:
                try:
                    return datetime.fromisoformat(iso_value)

                except ValueError:
                    return None

            case _:
                return None

    def with_creation(
        self,
        creation: datetime,
        /,
    ) -> Self:
        return self.__class__(
            {
                **self._values,
                "creation": creation.isoformat(),
            }
        )

    def merged_with(
        self,
        values: Self | MetaValues | None,
        /,
    ) -> Self:
        if not values:
            return self  # do not make a copy when nothing will be updated

        return self.__class__(
            {
                **self._values,  # already validated
                **{key: _validated_meta_value(value) for key, value in values.items()},
            }
        )

    def excluding(
        self,
        *excluded: str,
    ) -> Self:
        if not excluded:
            return self

        excluded_set: set[str] = set(excluded)
        return self.__class__(
            {key: value for key, value in self._values.items() if key not in excluded_set}
        )

    def updated(
        self,
        **values: MetaValue,
    ) -> Self:
        return self.__replace__(**values)

    def __replace__(
        self,
        **values: Any,
    ) -> Self:
        return self.merged_with(values)

    def __bool__(self) -> bool:
        return bool(self._values)

    def __contains__(
        self,
        element: Any,
    ) -> bool:
        return element in self._values

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )

    def __setitem__(
        self,
        key: str,
        value: Any,
    ) -> MetaValue:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" item - '{key}' cannot be modified"
        )

    def __delitem__(
        self,
        key: str,
    ) -> MetaValue:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" item - '{key}' cannot be deleted"
        )

    def __getitem__(
        self,
        key: str,
    ) -> MetaValue:
        return self._values[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self._values)

    def __len__(self) -> int:
        return len(self._values)

    def __copy__(self) -> Self:
        return self  # Metadata is immutable, no need to provide an actual copy

    def __deepcopy__(
        self,
        memo: dict[int, Any] | None,
    ) -> Self:
        return self  # Metadata is immutable, no need to provide an actual copy


def _validated_meta_value(value: Any) -> MetaValue:  # noqa: PLR0911
    match value:
        case None:
            return value

        case str():
            return value

        case int():
            return value

        case float():
            return value

        case bool():
            return value

        case [*values]:
            return tuple(_validated_meta_value(value) for value in values)

        case {**values}:
            return {key: _validated_meta_value(value) for key, value in values.items()}

        case other:
            raise TypeError(f"Invalid Meta value: {type(other)}")


META_EMPTY: Final[Meta] = Meta({})
